@@ -5,9 +5,11 @@ import pathlib
 import pickle
 import typing
 
+import natsort
 import networkx as nx
 import pandas
 import strictyaml as sy
+from ruamel.yaml import YAML
 
 
 class Category:
@@ -72,14 +74,7 @@ class Category:
         return code, spec
 
     def __str__(self) -> str:
-        s = "Category "
-        if len(self.codes) == 1:
-            s += f"{self.codes[0]}"
-        else:
-            s += f"{self.codes}"
-        if self.title is not None:
-            s += f" {self.title!r}"
-        return s
+        return f"{self.codes[0]} {self.title}"
 
     def __eq__(self, other: "Category"):
         if not isinstance(other, Category):
@@ -92,11 +87,18 @@ class Category:
         )
 
     def __repr__(self) -> str:
-        return f"<Category {self.codes[0]}>"
+        return f"<{self.categorization.name}: {self.codes[0]!r}>"
 
     def __hash__(self):
         return hash(self.categorization.name + self.codes[0])
 
+    def __lt__(self, other):
+        s = natsort.natsorted((self.codes[0], other.codes[0]))
+        if s[0] == self.codes[0] and not self == other:
+            return True
+        else:
+            return False
+
 
 class HierarchicalCategory(Category):
     """A single category from a HierarchicalCategorization."""
@@ -165,23 +167,6 @@ class HierarchicalCategory(Category):
         """
         return self.categorization.level(self)
 
-    def __str__(self) -> str:
-        s = "Category "
-        if len(self.codes) == 1:
-            s += f"{self.codes[0]}"
-        else:
-            s += f"{self.codes}"
-        if self.title is not None:
-            s += f" {self.title!r}"
-        s += (
-            f" children: "
-            f"{[tuple(sorted((c.codes[0] for c in cs))) for cs in self.children]!r}"
-        )
-        return s
-
-    def __repr__(self) -> str:
-        return f"<HierarchicalCategory {self.codes[0]}>"
-
 
 class Categorization:
     """A single categorization system.
@@ -225,6 +210,7 @@ class Categorization:
             "references": sy.Str(),
             "institution": sy.Str(),
             "last_update": sy.Str(),
+            "hierarchical": sy.Bool(),
             sy.Optional("version"): sy.Str(),
             "categories": sy.MapPattern(sy.Str(), Category._strictyaml_schema),
         }
@@ -263,15 +249,25 @@ class Categorization:
         self._add_categories(categories)
 
     @classmethod
-    def from_yaml(cls, filepath: typing.Union[str, pathlib.Path]) -> "Categorization":
+    def from_yaml(
+        cls, filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]]
+    ) -> "Categorization":
         """Read Categorization from a StrictYaml file."""
-        with open(filepath) as fd:
-            yaml = sy.load(fd.read(), schema=cls._strictyaml_schema)
+        try:
+            yaml = sy.load(filepath.read(), schema=cls._strictyaml_schema)
+        except AttributeError:
+            with open(filepath) as fd:
+                yaml = sy.load(fd.read(), schema=cls._strictyaml_schema)
         return cls.from_spec(yaml.data)
 
     @classmethod
     def from_spec(cls, spec: typing.Dict[str, typing.Any]) -> "Categorization":
         """Create Categorization from a Dictionary specification."""
+        if spec["hierarchical"] != cls.hierarchical:
+            raise ValueError(
+                "Specification is for a hierarchical categorization, use"
+                "HierarchicalCategorization.from_spec."
+            )
         last_update = datetime.date.fromisoformat(spec["last_update"])
         return cls(
             categories=spec["categories"],
@@ -284,15 +280,15 @@ class Categorization:
             version=spec.get("version", None),
         )
 
-    @classmethod
-    def from_pickle(cls, filepath: typing.Union[str, pathlib.Path]) -> "Categorization":
+    @staticmethod
+    def from_pickle(
+        filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]]
+    ) -> "Categorization":
         """De-serialize Categorization from a file written by to_pickle.
 
         Note that this uses the pickle module, which executes arbitrary code in the
         provided file. Only load from pickle files that you trust."""
-        with open(filepath, "rb") as fd:
-            spec = pickle.load(fd)
-        return cls.from_spec(spec)
+        return from_pickle(filepath)
 
     def to_spec(self) -> typing.Dict[str, typing.Any]:
         """Turn this categorization into a specification dictionary ready to be written
@@ -309,6 +305,7 @@ class Categorization:
             "comment": self.comment,
             "references": self.references,
             "institution": self.institution,
+            "hierarchical": self.hierarchical,
             "last_update": self.last_update.isoformat(),
         }
         if self.version is not None:
@@ -321,11 +318,12 @@ class Categorization:
         return spec
 
     def to_yaml(self, filepath: typing.Union[str, pathlib.Path]) -> None:
-        """Write to StrictYaml file."""
+        """Write to a YAML file."""
         spec = self.to_spec()
-        doc = sy.as_document(spec, self._strictyaml_schema)
+        yaml = YAML()
+        yaml.default_flow_style = False
         with open(filepath, "w") as fd:
-            fd.write(doc.as_yaml())
+            yaml.dump(spec, fd)
 
     def to_pickle(self, filepath: typing.Union[str, pathlib.Path]) -> None:
         """Serialize to a file using python's pickle."""
@@ -421,7 +419,8 @@ class Categorization:
         else:
             spec["last_update"] = last_update.isoformat()
 
-        spec["categories"].update(categories)
+        if categories is not None:
+            spec["categories"].update(categories)
 
         if alternative_codes is not None:
             for alias, primary in alternative_codes.items():
@@ -528,6 +527,7 @@ class HierarchicalCategorization(Categorization):
             "references": sy.Str(),
             "institution": sy.Str(),
             "last_update": sy.Str(),
+            "hierarchical": sy.Bool(),
             sy.Optional("version"): sy.Str(),
             "total_sum": sy.Bool(),
             sy.Optional("canonical_top_level_category"): sy.Str(),
@@ -605,22 +605,16 @@ class HierarchicalCategorization(Categorization):
         """Iterate over (primary code, category) pairs."""
         return self._primary_code_map.items()
 
-    @classmethod
-    def from_yaml(
-        cls, filepath: typing.Union[str, pathlib.Path]
-    ) -> "HierarchicalCategorization":
-        """Read HierarchicalCategorization from a StrictYaml file."""
-        with open(filepath) as fd:
-            yaml = sy.dirty_load(
-                fd.read(), allow_flow_style=True, schema=cls._strictyaml_schema
-            )
-        return cls.from_spec(yaml.data)
-
     @classmethod
     def from_spec(
         cls, spec: typing.Dict[str, typing.Any]
     ) -> "HierarchicalCategorization":
         """Create Categorization from a Dictionary specification."""
+        if spec["hierarchical"] != cls.hierarchical:
+            raise ValueError(
+                "Specification is for a non-hierarchical categorization, use"
+                "Categorization.from_spec."
+            )
         last_update = datetime.date.fromisoformat(spec["last_update"])
         return cls(
             categories=spec["categories"],
@@ -645,13 +639,14 @@ class HierarchicalCategorization(Categorization):
             Specification dictionary understood by `from_spec`.
         """
         # we can't call Categorization.to_spec here because we need to control ordering
-        # int the returned dict so that we get nicely ordered yaml files.
+        # in the returned dict so that we get nicely ordered yaml files.
         spec = {
             "name": self.name,
             "title": self.title,
             "comment": self.comment,
             "references": self.references,
             "institution": self.institution,
+            "hierarchical": self.hierarchical,
             "last_update": self.last_update.isoformat(),
         }
         if self.version is not None:
@@ -671,12 +666,130 @@ class HierarchicalCategorization(Categorization):
 
     @property
     def _canonical_subgraph(self) -> nx.DiGraph:
+        # TODO: from python 3.8 on, there is functools.cached_property to
+        # automatically cache this - as soon as we drop python 3.7 support, we can
+        # easily add it.
         return nx.DiGraph(
             self._graph.edge_subgraph(
                 ((u, v, 0) for (u, v, s) in self._graph.edges(data="set") if s == 0)
             )
         )
 
+    def _show_subtree(
+        self,
+        *,
+        node: HierarchicalCategory,
+        prefix="",
+        last=False,
+        format_func: typing.Callable[[HierarchicalCategory], str] = str,
+        maxdepth: typing.Optional[int],
+    ) -> str:
+        if maxdepth is not None:
+            maxdepth -= 1
+        if prefix:
+            if last:
+                r = f"{prefix[:-1]}╰{format_func(node)}\n"
+            else:
+                r = f"{prefix[:-1]}├{format_func(node)}\n"
+        else:
+            r = f"{format_func(node)}\n"
+
+        if maxdepth is not None and maxdepth == 0:
+            return r
+
+        child_sets = node.children
+        if len(child_sets) == 0:
+            return r
+        elif len(child_sets) == 1:
+            children = child_sets[0]
+            if children:
+                children_sorted = natsort.natsorted(children, key=format_func)
+                for child in children_sorted[:-1]:
+                    r += self._show_subtree(
+                        node=child,
+                        prefix=prefix + "│",
+                        format_func=format_func,
+                        maxdepth=maxdepth,
+                    )
+                r += self._show_subtree(
+                    node=children_sorted[-1],
+                    prefix=prefix + " ",
+                    last=True,
+                    format_func=format_func,
+                    maxdepth=maxdepth,
+                )
+        else:
+            prefix += "║"
+            first = True
+            for children in child_sets:
+                if children:
+                    if first:
+                        r += f"{prefix[:-1]}╠╤══\n"
+                        first = False
+                    else:
+                        r += f"{prefix[:-1]}╠╕\n"
+
+                    children_sorted = natsort.natsorted(children, key=format_func)
+                    for child in children_sorted[:-1]:
+                        r += self._show_subtree(
+                            node=child,
+                            prefix=prefix + "│",
+                            format_func=format_func,
+                            maxdepth=maxdepth,
+                        )
+                    r += self._show_subtree(
+                        node=children_sorted[-1],
+                        prefix=prefix + " ",
+                        last=True,
+                        format_func=format_func,
+                        maxdepth=maxdepth,
+                    )
+            r += f"{prefix[:-1]}╚═══\n"
+
+        return r
+
+    def show_as_tree(
+        self,
+        format_func: typing.Callable[[HierarchicalCategory], str] = str,
+        maxdepth: typing.Optional[int] = None,
+    ) -> str:
+        """Format the hierarchy as a tree.
+
+        Starting from top-level categories (i.e. categories without parents), the full
+        tree of categories is show, with children connected to their parents using
+        lines. If a parent category has one set of children, the children are connected
+        to each other and the parent with a simple line. If a parent category has
+        multiple sets of children, the sets are connected to parent with double lines
+        and the children in a set are connected to each other with simple lines.
+
+        Parameters
+        ----------
+        format_func: callable, optional
+            Function to call to format categories for display. Each category is
+            formatted for display using format_func(category), so format_func should
+            return a string without line breaks, otherwise the tree will look weird.
+            By default, str() is used, so that the first code and the title of the
+            category are used.
+        maxdepth: int, optional
+            Maximum depth to show in the tree. By default, goes to arbitrary depth.
+
+        Returns
+        -------
+        tree_str: str
+            Representation of the hierarchy as formatted string. print() it for optimal
+            viewing.
+        """
+        top_level_nodes = (node for node in self.values() if not node.parents)
+        r = ""
+        for top_level_node in top_level_nodes:
+            r += (
+                self._show_subtree(
+                    node=top_level_node, format_func=format_func, maxdepth=maxdepth
+                )
+                + "\n"
+            )
+        return r
+
     def extend(
         self,
         *,
@@ -828,3 +941,53 @@ class HierarchicalCategorization(Categorization):
 
         children = [set(children_dict[x]) for x in sorted(children_dict.keys())]
         return children
+
+
+def from_pickle(
+    filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]]
+) -> typing.Union[Categorization, HierarchicalCategorization]:
+    """De-serialize Categorization or HierarchicalCategorization from a file written by
+    to_pickle.
+
+    Note that this uses the pickle module, which executes arbitrary code in the
+    provided file. Only load from pickle files that you trust."""
+    try:
+        spec = pickle.load(filepath)
+    except TypeError:
+        with open(filepath, "rb") as fd:
+            spec = pickle.load(fd)
+
+    return from_spec(spec)
+
+
+def from_spec(
+    spec: typing.Dict[str, typing.Any]
+) -> typing.Union[Categorization, HierarchicalCategorization]:
+    """Create Categorization or HierarchicalCategorization from a dict specification."""
+    if spec["hierarchical"]:
+        return HierarchicalCategorization.from_spec(spec)
+    else:
+        return Categorization.from_spec(spec)
+
+
+def from_yaml(
+    filepath: typing.Union[str, pathlib.Path, typing.IO[bytes]]
+) -> typing.Union[Categorization, HierarchicalCategorization]:
+    """Read Categorization or HierarchicalCategorization from a StrictYaml file."""
+    try:
+        yaml = sy.load(filepath.read())
+        filepath.seek(0)
+    except AttributeError:
+        with open(filepath) as fd:
+            yaml = sy.load(fd.read())
+    hier = yaml.data["hierarchical"]
+    if hier in ("yes", "true", "True"):
+        cls = HierarchicalCategorization
+    elif hier in ("no", "false", "False"):
+        cls = Categorization
+    else:
+        raise ValueError(
+            f"'hierarchical' must be 'yes', 'true', 'True', 'no', 'false' or 'False',"
+            f" not {hier!r}."
+        )
+    return cls.from_yaml(filepath)
